"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const redis_1 = require("redis");
const client_1 = __importDefault(require("@node-redis/client/dist/lib/client"));
const ERRORS = {
    generic: (name, message) => {
        const error = new Error(message);
        error.name = name;
        error.message = message ?? "unknown";
        return error;
    },
    missingParameter: (item) => ERRORS.generic("missingParameter", `No ${item} supplied`),
    invalidFormat: (item) => ERRORS.generic("invalidFormat", `Invalid ${item} format`),
    cannotQuitExternalClient: () => ERRORS.generic("cannotQuitExternalClient", "Cannot quit external client"),
};
class RedisTagging {
    constructor(options = {}) {
        this.externalClient = false;
        this.VALID = { bucket: /^([a-zA-Z0-9_-]){1,80}$/ };
        this.redisns = (options.nsprefix || "rt") + ":";
        let port = options.port || 6379;
        let host = options.host || "127.0.0.1";
        if (options.client instanceof client_1.default) {
            this.redis = options.client;
            this.externalClient = true;
        }
        else {
            this.redis = (0, redis_1.createClient)(options.options ?? { socket: { port, host } });
            this.redis.on("error", err => console.log("Redis Client Error", err));
        }
    }
    quit() {
        if (this.externalClient) {
            throw ERRORS.cannotQuitExternalClient();
        }
        if (this.redis.isOpen)
            return this.redis.quit();
        return Promise.resolve();
    }
    async get(options) {
        const o = this.validate(options, ["bucket", "id"]);
        let ns = this.redisns + o.bucket;
        if (!this.redis.isOpen)
            await this.redis.connect();
        return this.redis.sMembers(`${ns}:ID:${o.id}`);
    }
    async set(options) {
        const o = this.validate(options, ["bucket", "id", "tags", "score"]);
        let ns = this.redisns + o.bucket;
        let id_index = `${ns}:ID:${o.id}`;
        const mc = await this.deleteID(ns, o.id);
        for (const tag of o.tags) {
            mc.push(["zincrby", ns + ":TAGCOUNT", 1, tag]);
            mc.push(["sadd", id_index, tag]);
            mc.push(["zadd", `${ns}:TAGS:${tag}`, o.score, o.id]);
        }
        if (o.tags.length) {
            mc.push(["sadd", ns + ":IDS", o.id]);
        }
        if (mc.length === 0)
            return true;
        if (!this.redis.isOpen)
            await this.redis.connect();
        const resp = await this.redis.multiExecutor(mc.map(v => ({ args: v.map(n => n.toString()) })));
        return true;
    }
    async remove(options) {
        const o = this.validate(options, ["bucket", "id"]);
        return this.set({ ...o, tags: [] });
    }
    async allids(options) {
        const o = this.validate(options, ["bucket"]);
        let ns = this.redisns + o.bucket;
        if (!this.redis.isOpen)
            await this.redis.connect();
        return this.redis.sMembers(ns + ":IDS");
    }
    async tags(options) {
        const o = this.validate(options, ["bucket", "tags", "offset", "limit", "withscores", "order", "type"]);
        let rndkey, resultkey, tagsresult;
        const ns = this.redisns + o.bucket;
        const prefix = ns + ":TAGS:";
        const lastelement = o.offset + o.limit - 1;
        const mc = [];
        if (!o.tags.length) {
            return {
                total_items: 0,
                items: [],
                limit: o.limit,
                offset: o.offset,
            };
        }
        if (o.tags.length > 1) {
            rndkey =
                ns +
                    new Date().getTime() +
                    "_" +
                    Math.floor(Math.random() * 9999999999);
            const _keys = o.tags.map((tag) => prefix + tag);
            mc.push([`z${o.type}store`, rndkey, _keys.length]
                .concat(_keys)
                .concat(["AGGREGATE", "MIN"]));
            if (o.limit > 0) {
                resultkey = rndkey;
            }
        }
        else if (o.tags.length === 1) {
            mc.push(["zcard", prefix + o.tags[0]]);
            if (o.limit > 0) {
                resultkey = prefix + o.tags[0];
            }
        }
        if (o.limit > 0) {
            tagsresult = [
                `z${o.order}range`,
                resultkey,
                o.offset,
                lastelement,
            ];
            if (o.withscores) {
                tagsresult.push("WITHSCORES");
            }
            mc.push(tagsresult);
        }
        if (o.tags.length > 1) {
            mc.push(["del", rndkey]);
        }
        if (!this.redis.isOpen)
            await this.redis.connect();
        const resp = await this.redis.multiExecutor(mc.map(v => ({ args: v.map(n => n.toString()) })));
        let rows;
        if (o.limit === 0) {
            rows = [];
        }
        else {
            rows = resp[1];
        }
        if (rows.length && o.withscores) {
            rows = rows.reduce((acc, curr, i, a) => {
                if (i % 2 === 0) {
                    acc.push({
                        id: curr,
                        score: +a[i + 1],
                    });
                }
                return acc;
            }, []);
        }
        return {
            total_items: resp[0] ? +resp[0] : 0,
            items: rows,
            limit: o.limit,
            offset: o.offset,
        };
    }
    async toptags(options) {
        const o = this.validate(options, ["bucket", "amount"]);
        const ns = this.redisns + o.bucket;
        o.amount = o.amount - 1;
        const rediskey = ns + ":TAGCOUNT";
        const mc = [
            ["zcard", rediskey],
            ["zrevrange", rediskey, 0, o.amount, "WITHSCORES"],
        ];
        if (!this.redis.isOpen)
            await this.redis.connect();
        const resp = await this.redis.multiExecutor(mc.map(v => ({
            args: v.map(n => n.toString())
        })));
        const rows = resp[1].reduce((acc, curr, i, a) => {
            if (i % 2 === 0) {
                acc.push({
                    tag: curr,
                    count: +a[i + 1],
                });
            }
            return acc;
        }, []);
        return {
            total_items: resp[0] ? +resp[0] : 0,
            items: rows,
        };
    }
    async buckets() {
        if (!this.redis.isOpen)
            await this.redis.connect();
        const resp = await this.redis.keys(this.redisns + "*:TAGCOUNT");
        return resp.map((v) => v.split(":")[1]);
    }
    async removebucket(options) {
        const o = this.validate(options, ["bucket"]);
        const ns = this.redisns + o.bucket;
        const mc = [
            ["smembers", ns + ":IDS"],
            ["zrange", ns + ":TAGCOUNT", 0, -1],
        ];
        if (!this.redis.isOpen)
            await this.redis.connect();
        const resp = await this.redis.multiExecutor(mc.map(v => ({ args: v.map(n => n.toString()) })));
        const rkeys = [ns + ":IDS", ns + ":TAGCOUNT"];
        for (const e of resp[0]) {
            rkeys.push(ns + ":ID:" + e);
        }
        for (const e of resp[1]) {
            rkeys.push(ns + ":TAGS:" + e);
        }
        if (!this.redis.isOpen)
            await this.redis.connect();
        await this.redis.del(rkeys);
        return true;
    }
    async deleteID(ns, id) {
        const mc = [];
        const id_index = ns + ":ID:" + id;
        if (!this.redis.isOpen)
            await this.redis.connect();
        const resp = await this.redis.sMembers(id_index);
        if (resp.length) {
            for (const tag of resp) {
                mc.push(["zincrby", ns + ":TAGCOUNT", -1, tag]);
                mc.push(["zrem", `${ns}:TAGS:${tag}`, id]);
            }
            mc.push(["del", id_index]);
            mc.push(["srem", ns + ":IDS", id]);
            mc.push(["zremrangebyscore", ns + ":TAGCOUNT", 0, 0]);
        }
        return mc;
    }
    validate(options, keys) {
        const validOptions = {};
        for (const key of keys) {
            switch (key) {
                case "bucket":
                    if (!options[key])
                        throw ERRORS.missingParameter(key);
                    validOptions[key] = options[key].toString();
                    if (!this.VALID[key].test(validOptions[key]))
                        throw ERRORS.invalidFormat(key);
                    break;
                case "id":
                    if (!options[key])
                        throw ERRORS.missingParameter(key);
                    validOptions[key] = options[key].toString();
                    if (!validOptions[key].length)
                        throw ERRORS.missingParameter(key);
                    break;
                case "tags":
                    if (!options[key])
                        throw ERRORS.missingParameter(key);
                    if (!Array.isArray(options[key])) {
                        throw ERRORS.invalidFormat(key);
                    }
                    validOptions[key] = options[key].map(v => v.toString());
                    break;
                case "score":
                    validOptions[key] = options[key] ? options[key] : 0;
                    if (typeof validOptions[key] === "string")
                        validOptions[key] = parseInt(options[key], 10);
                    if (isNaN(validOptions[key]))
                        throw ERRORS.invalidFormat(key);
                    break;
                case "limit":
                    if (isNaN(options[key]))
                        validOptions[key] = 100;
                    else
                        validOptions[key] = Math.abs(options[key]);
                    break;
                case "offset":
                case "withscores":
                case "amount":
                    validOptions[key] = options[key] ? options[key] : 0;
                    if (typeof validOptions[key] === "string")
                        validOptions[key] = parseInt(validOptions[key], 10);
                    if (isNaN(validOptions[key]))
                        throw ERRORS.invalidFormat(key);
                    validOptions[key] = Math.abs(validOptions[key]);
                    break;
                case "order":
                    validOptions[key] = options[key] === "asc" ? "" : "rev";
                    break;
                case "type":
                    if (options[key]?.toLowerCase() === "union")
                        validOptions[key] = "union";
                    else
                        validOptions[key] = "inter";
                    break;
                default:
                    throw ERRORS.invalidFormat(key);
            }
        }
        return validOptions;
    }
}
exports.default = RedisTagging;
